/*  Copyright 2010-2011 Guillaume Duhamel

    This file is part of Gophrier.

    Gophrier is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Gophrier is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Gophrier; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
*/

#include "request.h"
#include "conf.h"
#include "dir.h"
#include "file.h"
#include "error.h"
#include "exe.h"
#include "map.h"
#include "dbstring.h"
#include "sock.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/socket.h>

static char * request_path_cleanup(const char * inpath, char ** query);
static char * request_query_decode(const char * query);

Request * request_new(Socket * sock, const char * path)
{
    Request * request;
    char * rootdir;

    request = (Request *) malloc(sizeof(Request));
    request->sock = sock;

    request->path = request_path_cleanup(path, &request->query);

    rootdir = conf_get_root_dir();
    request->fullpath = malloc(strlen(request->path) + strlen(rootdir) + 1);
    sprintf(request->fullpath, "%s%s", rootdir, request->path);

    return request;
}

void request_exec(Request * request)
{
    struct stat info;

    if (-1 == stat(request->fullpath, &info))
        return error_parse(request);
    if (S_ISDIR(info.st_mode))
    {
        char gmpath[1024];
        struct stat gminfo;
        sprintf(gmpath, "%s/gophermap", request->fullpath);
        if (-1 == stat(gmpath, &gminfo))
            dir_parse(request);
        else
        {
            free(request->fullpath);
            request->fullpath = strdup(gmpath);
            map_parse(request);
        }
    }
    else if (info.st_mode & S_IXOTH)
        exe_parse(request, &info);
    else
        file_parse(request);
}

void request_free(Request * request)
{
    free(request->fullpath);
    free(request->path);
    free(request->query);
    free(request);
}

static char * request_path_cleanup(const char * inpath, char ** query)
{
    char * outpath;
    char * slash;
    char * path = strdup(inpath);
    char * current = path;
    int toskip = 0;
    DBString * string;
    int inlen = strlen(inpath);
    char * querymark;

    querymark = strrchr(path, '?');
    if (querymark)
    {
        *query = request_query_decode(querymark + 1);
        *querymark = 0;
    }
    else
        *query = NULL;

    string = dbstring_new(inlen > 1 ? inlen : 1);

    while(NULL != (slash = strrchr(current, '/')))
    {
        if (strcmp("/", slash) == 0)
            ;
        else if (strcmp("..", slash + 1) == 0)
            toskip++;
        else if (toskip)
            toskip--;
        else
            dbstring_prepend(string, slash);
        *slash = 0;
    }
    dbstring_prepend(string, current);
    free(path);

    if (strcmp("/", dbstring_content(string)) == 0)
        dbstring_shift(string, 1);

    outpath = strdup(dbstring_content(string));
    dbstring_free(string);

    return outpath;
}

ssize_t request_send(Request * request, const void * buf, size_t len)
{
    return send(request->sock->fd, buf, len, MSG_NOSIGNAL);
}

static char * request_query_decode(const char * query)
{
    char * decoded = malloc(strlen(query) + 1);
    const char * c = query;
    int i = 0;

    while(*c)
    {
        if (*c == '%')
        {
            unsigned int code;

            sscanf(c + 1, "%2X", &code);
            decoded[i] = code;
            c += 2;
        }
        else
            decoded[i] = *c;
        c++;
        i++;
    }
    decoded[i] = 0;

    return decoded;
}
